// Adapted from google-cloud-profiler.
// Copyright 2018 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef MPROFILE_SRC_SCOPED_OBJECT_H_
#define MPROFILE_SRC_SCOPED_OBJECT_H_

#include <Python.h>

#include <memory>

struct PyObjectDecReffer {
  void operator()(PyObject* py_object) const { Py_XDECREF(py_object); }
};

typedef std::unique_ptr<PyObject, PyObjectDecReffer> PyObjectRef;

#endif  // MPROFILE_SRC_SCOPED_OBJECT_H_
